/*
 *	Copyright (C) Jan Bobrowski <jb@wizard.ae.krakow.pl>
 *
 *	This library is free software; you can redistribute and modify
 *	it under the terms of the GNU Lesser General Public License
 *	version 2.1 as published by the Free Software Foundation.
 */

//#define NDEBUG
#include <string.h>
#include <unistd.h>
#include <assert.h>
#include "kbd.h"

//#define ALLOW_DEL_CHG 1
#define WANT_MOUSE
#define WANT_UNICODE

#ifdef WANT_UNICODE
int getkey_utf8 = 0;
#endif

#define elemof(T) (sizeof T/sizeof*T)
#define endof(T) (T + elemof(T))

typedef unsigned char u8;

static int csitbl[] = {
/* @ */	kINS, kUP, kDOWN, kRIGHT, kLEFT, kCENTER, kEND,
/* G */	kPGDN, kHOME, kPGUP, kWINL, kDEL, kINS,
/* M */	kF1, kF2, kF3, kF4, kF5, kF6, kF7, kF8, kF9, kF10, kF11, kF12,
/* Y */	kHOME, kSHIFT|kTAB, 0,0,0,0,0,0,
/* a */	kSHIFT|kUP, kSHIFT|kDOWN, kSHIFT|kRIGHT, kSHIFT|kLEFT,
/* e */	kPGDN, kSHIFT|kEND, 0, kSHIFT|kHOME, 0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0, 0,0, kMENU, kWINR
};

/*static int ss3tbl[] = {
	0, kUP, kDOWN, kRIGHT, kLEFT, 0,0,0,0, kTAB ,0,0,0, kENTER,
	0,0, kF1, kF2, kF3, kF4, 0,0,0,0, '=',0,0, 0,0,0,0,0,0,
	kCTRL|kUP, kCTRL|kDOWN, kCTRL|kRIGHT, kCTRL|kLEFT,
	0,0,0,0,0, '*', '+', ',', '-', '.', '/', '0','1','2','3','4','5','6','7','8','9'
};*/

static int ss3tbl[] = {
	0, kUP, kDOWN, kRIGHT, kLEFT, 0,0,0,0, kTAB ,0,0,0, kENTER,
	0,0, kF1, kF2, kF3, kF4, 0,0,0,0, '=',0,0, 0,0,0,0,0,0,
	kCTRL|kUP, kCTRL|kDOWN, kCTRL|kRIGHT, kCTRL|kLEFT,
	0,0,0,0,0, '*', '+', ',', '-', kDEL, '/', kINS, kEND, kDOWN,
	kPGDN, kLEFT, kCENTER, kRIGHT, kHOME, kUP, kPGUP
};

static int csi_tilde[] = {
	0, kHOME, kINS, kDEL, kEND, kPGUP, kPGDN,
	kHOME, kEND, kF3, kF4, kF1, kF2, kF3, kF4, kF5,
	kF5, kF6, kF7, kF8, kF9, kF10, 0, kF11, kF12,
	kF13, kF14, 0, kF15, kF16, 0, kF17, kF18, kF19, kF20
};

#define CSI_Q_BASE 139
static int csi_q[] = {
	kINS, 0,0,0, kSHIFT|kHOME, 0,0, kEND, 0,0,0, kPGUP, kSCRDN,
	0,0, kPGDN, kSCRUP, 0,0, kSHIFT|kLEFT, kSHIFT|kRIGHT
};

#define CSI_Z_BASE 192
static int csi_z[] = {
	kF11, kF12, 0,0,0,0,0,0, 0,0,0,0,0, 0,0,0,0,0, 0,0,0,0,
	kHOME, 0, kPGUP, 0, kCENTER, 0, kEND, 0, kPGDN, 0,
	kF1, kF2, kF3, kF4, kF5, kF6, kF7, kF8, kF9,
	kF10, kF11, kF12, 0,0,0,0,0, 0,0,0,0,0, 0, kINS, 0, kDEL
};

static int st_start(int c);
#ifdef WANT_UNICODE
static int st_utf8(int c);
#endif
static int st_esc(int c);
static int st_csiss3(int c);
static int st_arg(int c);
static int st_mouse(int c);

static int (*state)(int) = st_start;

static struct seq {
	u8 type, mod;
	u8 argf:1, dblesc:1;
#if ALLOW_DEL_CHG
	int del;
#endif
	int arg[16], *argp;
} seq = {
	argp: seq.arg,
#if ALLOW_DEL_CHG
	del: kBS
#define DEL_IS(K) (seq.del=(K))
#define KEY_OF_DEL seq.del
#else
#define DEL_IS(K) 
#define KEY_OF_DEL kBS
#endif
};

#ifdef WANT_MOUSE
static struct mouse_event mev;
#endif

static int lowtbl[32] = {
	kCTRL|'2',0,0,0,0,0,0,0, kBS, kTAB, kSHIFT|kENTER, 0,0,
	kENTER, 0,0,0,0,0,0,0,0,0,0,0,0,0, kESC, 0,
	kCTRL|'5', kCTRL|'6', kCTRL|'7'
};

static int ordkey(int c)
{
	int k = c;
	if(c < 32) {
		if(c == 8) DEL_IS(kDEL); /* DEL is DEL */
		k = lowtbl[c];
	} else if(c >= 127 && c < 0xA0) {
		k = kNOKEY;
		if(c == 127)
			k = KEY_OF_DEL;
	}
	return k;
}

static int st_start(int c)
{
	int k;
	if(c == 27 || c == 0x9B || c == 0x8F) {
		seq.type = seq.mod = 0;
		seq.argp = seq.arg;
		seq.arg[0] = 0;
		seq.arg[1] = 0;
		seq.argf = seq.dblesc = 0;

		state = st_esc;
		if(c & 0x80) {
			seq.type = c - 64;
			state = st_csiss3;
		}
		return kMORE;
	}
#ifdef WANT_UNICODE
	if(getkey_utf8 && c >= 0x80) {
		int b;
		if(c<0xC2 || c>0xF4)
			return kERROR;
		if(c < 0xE0) b = 6, c &= 0x1F;
		else if(c < 0xF0) b = 12, c &= 0x0F;
		else b = 18, c &= 0x07;
		seq.arg[0] = c << b;
		seq.arg[1] = b;
		state = st_utf8;
		return kMORE;
	}
#endif
	k = ordkey(c);
	if(!k) {
		int m = kCTRL;
		c ^= 64;
		if(c >= 128) {
			c ^= 192;
			m ^= kALT | kCTRL;
		}
		k = ordkey(c);
		if(k) k |= m;
	}
	return k;
}

#ifdef WANT_UNICODE
static int st_utf8(int c) {
	if((c&0xC0) != 0x80) {
		state = st_start;
		return st_start(c);
	}
	seq.arg[1] -= 6;
	seq.arg[0] |= (c & 0x3F)<<seq.arg[1];
	if(seq.arg[1])
		return kMORE;
	state = st_start;
	return seq.arg[0];
}
#endif

static int st_esc(int c)
{
	if(c == kMORE)
		return kINCOMPLETE;
	if(c == '[' || c == 'O') {
		seq.type = c;
		state = st_csiss3;
		return kMORE;
	}
	if(c == 27) {
		if(seq.dblesc)
			return kINCOMPLETE;
		seq.dblesc = 1;
		return kMORE;
	}

	c = ordkey(c);
	if(c) c |= kALT;
	return c;
}

#define CSISEQ (seq.type == '[')
#define SS3SEQ (seq.type == 'O')

static int csiss3(int c)
{
	int m = seq.arg[0], v = 0;
	unsigned a = seq.arg[0];

	if(SS3SEQ) {
		c -= 0x40;
		if(c < elemof(ss3tbl))
			v = ss3tbl[c];
		goto ready;
	}
	if(seq.mod == '[') {
		c -= 'A';
		if(c < 12)
			v = kF1 + c;
		goto ready;
	}
	if(seq.mod == '?') {
		if(c == 'c') {
			v = kDECID;
			goto inclarg;
		}
		goto ret;
	}
	if(seq.argf && (c == '~' || c == '$' || c == '^' || c == '@')) {
		if(a >= elemof(csi_tilde))
			goto ret;
		v = csi_tilde[a];
		if(!v) goto ret;
		if(v == kDEL) DEL_IS(kBS); /* DEL is BS */
		if(c != '~') {
			if(c != '^') v |= kSHIFT;
			if(c != '$') v |= kCTRL;
		}
		m = seq.arg[1];
		goto ready;
	}
	if(c == 'q') {
		m = seq.arg[1];
		if(a < CSI_Q_BASE) goto fkey;
		a -= CSI_Q_BASE;
		if(a >= elemof(csi_q)) goto ret;
		v = csi_q[a];
		goto ready;
	}
	if(c == 's') {
		m = seq.arg[1];
fkey:
		if(--a >= 12) goto ret;
		v = kF1 + a;
		goto ready;
	}
	if(c == 'z') {
		m = seq.arg[1];
		if(a < elemof(csi_tilde)) {
			v = csi_tilde[a];
			goto ready;
		}
		a -= CSI_Z_BASE;
		if(a >= elemof(csi_z))
			goto ret;
		v = csi_z[a];
		goto ready;
	}
	if(c == 'R' && seq.argp == &seq.arg[1]) {
		v = kCPOS;
inclarg:
		v |= seq.arg[0]<<16 | seq.arg[1]<<24;
		goto ret;
	}
	if(c == 't') {
		v = kTERM_T;
		v |= seq.arg[0] & 0xFF;
		v |= seq.arg[1] << 16;
		if(seq.argp > seq.arg + 1)
			v |= seq.arg[2] << 24;
		goto ret;
	}

	c -= 0x40;
	if(c < elemof(csitbl))
		v = csitbl[c];
ready:
	if(v) {
		if(m) {
			m = (m-1 & 7) << 12;	/*XXX Mod==9 */
			v = v&kMASK | m;
		}
		if(seq.dblesc) v |= kALT;
	}
ret:
	return v;
}

static int st_csiss3(int c)
{
	if(c == kMORE || c < ' ')
		return kINCOMPLETE;
	if((c == '[' || c == '?') && CSISEQ) {
		if(seq.mod) return kINCOMPLETE;
		seq.mod = c;
		return kMORE;
	}
	if(c >= '0' && c <= '9') {
		seq.arg[0] = c - '0';
		goto goarg;
	}
	if(c == ';') {
		*++seq.argp = 0;
goarg:
		seq.argf = 1;
		state = st_arg;
		return kMORE;
	}
	if(c == 'M' && CSISEQ) {
		state = st_mouse;
		return kMORE;
	}
	return csiss3(c);
}

static int st_arg(int c)
{
	if(c == kMORE)
		return kINCOMPLETE;
	if(c >= '0' && c <= '9') {
		*seq.argp = 10**seq.argp + c-'0';
		return kMORE;
	}
	if(c == ';') {
		if(++seq.argp >= endof(seq.arg))
			return kNOKEY;
		*seq.argp = 0;
		return kMORE;
	}

	return csiss3(c);
}

#ifdef WANT_MOUSE
static inline int xterm_mouse()
{
	static u8 m;
	static u8 butt[] = {1,4,2,0};
	int v = seq.arg[0] & 0xE3;
	if(m&0xC0 && (v&3) == 3) {
		m = seq.arg[0];
		return kNOKEY;
	}
	m = seq.arg[0];
	mev.state &= 7;
	if(v & 0xC0) {
		mev.code = kmWHEEL;
		mev.wy = v&1 ? 1 : -1;
		mev.button = 0;
		goto end;
	}
	mev.wy = 0;
	mev.button = butt[v&3];
	if(v & 0x20) {
		mev.code = kmDRAG;
		goto end;
	}
	v &= 3;
	if(v<3) {
		mev.code = kmPRESS;
//		if(mev.state & b)
//			mev.code = kmDRAG;
		mev.state |= mev.button;
	} else if(v==3) {
		mev.code = kmRELEASE;
		if(!mev.state)
			mev.code = kmMOVE;
		mev.button = mev.state&7;
		mev.state = 0;
	}

end:
	v = (seq.arg[0]&0x1C) << 10;
	mev.state |= v;
	mev.x = seq.arg[1] - 1;
	mev.y = seq.arg[2] - 1;
	return kMOUSE | v;
}	
#endif

static int st_mouse(int c)
{
	if(c == kMORE)
		return kINCOMPLETE;
	*seq.argp++ = c - 0x20;
	if(seq.argp < seq.arg + 3)
		return kMORE;

#ifdef WANT_MOUSE
	return xterm_mouse();
#else
	return kNOKEY;
#endif
}

static int waitdata;
static u8 buf[20], *p = buf, *e = buf;

static int incomplete()
{
	int l = e - buf;
	p = buf;

	if(l == 1)
		return ordkey(*p++);
	if(l < sizeof(buf)) buf[l] = 0;
	if(!memcmp(p, "\33[M", 3)) {
		p += 3;
		return kF1;
	} else if(!memcmp(p, "\233M", 2)) {
		p += 2;
		return kF1;
	}
	if(*p++ == 27)
		return ordkey(*p++) | kALT;
	return ordkey(p[-1]);
}

int getkey()
{
	int v, c;

again:
	if(p == e) {
		int l = endof(buf) - p;
		if(l == 0) goto nomore;
		v = read(0, p, l);
		if(v <= 0) {
			if(waitdata) {
nomore:
				c = kMORE;
				goto key;
			}
			return v<0 ? kNOKEY : kERROR;
		}
		e = p + v;
	}
	c = *p++;
key:
	v = state(c);
	if(v == kINCOMPLETE)
		v = incomplete();
	assert(c >= 0 || v >= 0);
	if(v < 0) {
		if(p < e)
			goto again;
		waitdata = 1;
		return kMORE;
	}
	waitdata = 0;

	{
		u8 *q = buf;
		while(p < e) *q++ = *p++;
		e = q;
		p = buf;
	}

	state = st_start;
	if(!(v&kMASK)) goto again;

	return v;
}

#ifdef WANT_MOUSE
const struct mouse_event *getmouse()
{
	return &mev;
}
#endif
